from __future__ import annotations

import json
import numpy as np
import pandas as pd

from operator import add
from enum import Enum
from typing import Optional, Dict, List, Literal, Union
from typing import TYPE_CHECKING, Callable
from pydantic_core import CoreSchema
from pydantic import (
    BaseModel,
    ConfigDict,
    GetJsonSchemaHandler,
    model_validator,
    field_validator,
    ValidationError,
)
from pydantic.json_schema import JsonSchemaValue

import ITR

from ITR.data.osc_units import (
    ureg,
    Q_,
    M_,
    PA_,
    BenchmarkMetric,
    BenchmarkQuantity,
    ProductionMetric,
    ProductionQuantity,
    EmissionsMetric,
    EmissionsQuantity,
    EI_Metric,
    EI_Quantity,
    MonetaryQuantity,
    Quantity_type,
)
from ITR.configs import ProjectionControls, LoggingConfig

import logging

logger = logging.getLogger(__name__)
LoggingConfig.add_config_to_logger(logger)

import pint
from pint.errors import DimensionalityError
from pint_pandas import PintType
from pint_pandas.pint_array import PintSeriesAccessor


class SortableEnum(Enum):
    def __str__(self):
        return self.name

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            order = list(self.__class__)
            return order.index(self) >= order.index(other)
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            order = list(self.__class__)
            return order.index(self) > order.index(other)
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            order = list(self.__class__)
            return order.index(self) <= order.index(other)
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            order = list(self.__class__)
            return order.index(self) < order.index(other)
        return NotImplemented


class EScope(SortableEnum):
    S1 = "S1"
    S2 = "S2"
    S3 = "S3"
    S1S2 = "S1+S2"
    S1S2S3 = "S1+S2+S3"
    AnyScope = "AnyScope"

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()

    @classmethod
    def get_scopes(cls) -> List[str]:
        """
        Get a list of all scopes.
        :return: A list of EScope string values
        """
        return ["S1", "S2", "S3", "S1S2", "S1S2S3"]

    @classmethod
    def get_result_scopes(cls) -> List["EScope"]:
        """
        Get a list of scopes that should be calculated if the user leaves it open.

        :return: A list of EScope objects
        """
        # FIXME: Should this also contain cls.S2 or no?
        return [cls.S1, cls.S1S2, cls.S3, cls.S1S2S3]


class ETimeFrames(SortableEnum):
    """
    TODO: add support for multiple timeframes. Long currently corresponds to 2050.
    """

    SHORT = "short"
    MID = "mid"
    LONG = "long"


class ECarbonBudgetScenario(Enum):
    P25 = "25 percentile"
    P75 = "75 percentile"
    MEAN = "Average"


class EScoreResultType(Enum):
    DEFAULT = "Default"
    TRAJECTORY_ONLY = "Trajectory only"
    TARGET_ONLY = "Target only"
    COMPLETE = "Complete"

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()

    @classmethod
    def get_result_types(cls) -> List[str]:
        """
        Get a list of all result types, ordered by priority (first << last priority).
        :return: A list of the EScoreResultType values
        """
        return [
            EScoreResultType.DEFAULT,
            EScoreResultType.TRAJECTORY_ONLY,
            EScoreResultType.TARGET_ONLY,
            EScoreResultType.COMPLETE,
        ]


class AggregationContribution(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    company_name: str
    company_id: str
    temperature_score: Quantity_type("delta_degC")
    contribution_relative: Optional[Quantity_type("percent")] = None
    contribution: Optional[Quantity_type("delta_degC")] = None

    def __getitem__(self, item):
        return getattr(self, item)


class Aggregation(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    score: Quantity_type("delta_degC")
    # proportion is a number from 0..1
    proportion: float
    contributions: List[AggregationContribution]

    def __getitem__(self, item):
        return getattr(self, item)


class ScoreAggregation(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    all: Aggregation
    influence_percentage: Quantity_type("percent")
    grouped: Dict[str, Aggregation]

    def __getitem__(self, item):
        return getattr(self, item)


class ScoreAggregationScopes(BaseModel):
    S1: Optional[ScoreAggregation] = None
    S2: Optional[ScoreAggregation] = None
    S1S2: Optional[ScoreAggregation] = None
    S3: Optional[ScoreAggregation] = None
    S1S2S3: Optional[ScoreAggregation] = None

    def __getitem__(self, item):
        return getattr(self, item)


class ScoreAggregations(BaseModel):
    short: Optional[ScoreAggregationScopes] = None
    mid: Optional[ScoreAggregationScopes] = None
    long: Optional[ScoreAggregationScopes] = None

    def __getitem__(self, item):
        return getattr(self, item)


class PortfolioCompany(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    company_name: str
    company_id: str
    company_isin: Optional[str] = None
    investment_value: MonetaryQuantity
    user_fields: Optional[dict] = None


# U is Unquantified, which is presently how our benchmarks come in (production_metric comes in elsewhere)
class UProjection(BaseModel):
    year: int
    value: float


class IProjection(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    year: int
    value: BenchmarkQuantity


class IBenchmark(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    sector: str
    region: str
    benchmark_metric: BenchmarkMetric
    projections_nounits: Optional[List[UProjection]] = None
    projections: Optional[List[IProjection]] = None
    base_year_production: Optional[ProductionQuantity] = None  # FIXME: applies only to production benchmarks

    def __init__(
        self,
        benchmark_metric,
        projections_nounits=None,
        projections=None,
        base_year_production=None,
        *args,
        **kwargs,
    ):
        # FIXME: Probably want to define `target_end_year` to be 2051, not 2050...
        super().__init__(
            benchmark_metric=benchmark_metric,
            projections_nounits=projections_nounits,
            projections=projections,
            base_year_production=base_year_production,
            *args,
            **kwargs,
        )
        # Sadly we need to build the full projection range before cutting it down to size...
        # ...until Tiemann learns the bi-valence of dict and Model parameters
        if self.projections_nounits:
            if self.projections:
                # Check if we've already seen/processed these exact projections
                changed_projections = [
                    p
                    for p in self.projections
                    if not any([n for n in self.projections_nounits if n.year == p.year and n.value == p.value.m])
                ]
                if changed_projections:
                    raise ValueError
                return
            self.projections = [
                IProjection(year=p.year, value=BenchmarkQuantity(Q_(p.value, benchmark_metric)))
                for p in self.projections_nounits
                if p.year in range(ProjectionControls.BASE_YEAR, ProjectionControls.TARGET_YEAR + 1)
            ]
        elif not self.projections:
            logger.warning(f"Empty Benchmark for sector {sector}, region {region}")

    def __getitem__(self, item):
        return getattr(self, item)


class IBenchmarks(BaseModel):
    benchmarks: List[IBenchmark]
    production_centric: bool = False

    def __getitem__(self, item):
        return getattr(self, item)


# These IProductionBenchmarkScopes and IEIBenchmarkScopes are vessels for holding initialization data
# The CompanyDataProvider methods create their own dataframes that are then used throughout


class IProductionBenchmarkScopes(BaseModel):
    AnyScope: Optional[IBenchmarks] = None
    S1: Optional[IBenchmarks] = None
    S2: Optional[IBenchmarks] = None
    S1S2: Optional[IBenchmarks] = None
    S3: Optional[IBenchmarks] = None
    S1S2S3: Optional[IBenchmarks] = None

    def __getitem__(self, item):
        return getattr(self, item)


class IEIBenchmarkScopes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    S1: Optional[IBenchmarks] = None
    S2: Optional[IBenchmarks] = None
    S1S2: Optional[IBenchmarks] = None
    S3: Optional[IBenchmarks] = None
    S1S2S3: Optional[IBenchmarks] = None
    benchmark_temperature: Quantity_type("delta_degC")
    benchmark_global_budget: Quantity_type("Gt CO2")
    is_AFOLU_included: bool

    def __getitem__(self, item):
        return getattr(self, item)


class ICompanyEIProjection(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    year: int
    value: Optional[EI_Quantity] = None

    def __getitem__(self, item):
        return getattr(self, item)

    def __eq__(self, o):
        if self.year != o.year:
            raise ValueError(f"EI Projection years not aligned for __eq__(): {self.year} vs. {o.year}")
        if ITR.isna(self.value.m) and ITR.isna(o.value.m):
            return True
        return self.value == o.value

    def add(self, o):
        if self.year != o.year:
            # breakpoint()
            raise ValueError(f"EI Projection years not aligned for add(): {self.year} vs. {o.year}")
        return ICompanyEIProjection(
            year=self.year,
            value=self.value if ITR.isna(o.value.m) else self.value + o.value.to(self.value.units),
        )

    def min(self, o):
        if self.year != o.year:
            raise ValueError(f"EI Projection years not aligned for min(): {self.year} vs. {o.year}")
        return ICompanyEIProjection(year=self.year, value=min(self.value, o.value))


class ICompanyEIProjections(BaseModel):
    ei_metric: EI_Metric
    projections: List[ICompanyEIProjection]

    def __getitem__(self, item):
        return getattr(self, item)

    def __str__(self):
        # Work-around for https://github.com/hgrecco/pint/issues/1687
        ei_metric = ureg.parse_units(self.ei_metric)
        series = (
            lambda z: (
                idx := z[0],
                values := z[1],
                pd.Series(PA_(np.asarray(values), dtype=str(ei_metric)), index=idx),
            )[-1]
        )(list(zip(*[(x.year, round(ITR.Q_m_as(x.value, ei_metric), 4)) for x in self.projections])))
        return str(series)


class DF_ICompanyEIProjections(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    ei_metric: Optional[EI_Metric] = None
    projections: pd.Series

    @field_validator("projections")
    def val_projections(cls, v: pd.Series):
        if isinstance(v.pint, PintSeriesAccessor):
            return v
        raise ValidationError(f"{v} is not composed of a PintArray")

    def __init__(self, icompany_ei_projections=None, *args, **kwargs):
        projections = None
        projections_gen = None
        if icompany_ei_projections is not None:
            ei_metric = icompany_ei_projections.ei_metric
            projections_gen = icompany_ei_projections.projections
        else:
            ei_metric = kwargs["ei_metric"]
            projections = kwargs["projections"]
            if not isinstance(projections, pd.Series):
                projections_gen = projections
                projections = None
        if projections_gen is not None:
            # Work-around for https://github.com/hgrecco/pint/issues/1687
            ei_metric = ureg.parse_units(ei_metric)
            years, values = list(
                map(
                    list,
                    zip(
                        *[
                            (
                                x["year"],
                                np.nan if x["value"] is None else ITR.Q_m_as(x["value"], ei_metric, inplace=True),
                            )
                            for x in projections_gen
                        ]
                    ),
                )
            )
            projections = pd.Series(
                PA_(np.asarray(values), dtype=str(ei_metric)),
                index=pd.Index(years, name="year"),
                name="value",
            )
        super().__init__(ei_metric=str(ei_metric), projections=projections)


class ICompanyEIProjectionsScopes(BaseModel):
    S1: Optional[DF_ICompanyEIProjections] = None
    S2: Optional[DF_ICompanyEIProjections] = None
    S1S2: Optional[DF_ICompanyEIProjections] = None
    S3: Optional[DF_ICompanyEIProjections] = None
    S1S2S3: Optional[DF_ICompanyEIProjections] = None

    def __init__(self, *args, **kwargs):
        # We don't validate anything in the first step because incoming parameters are the wild west
        # (dict, ICompanyEIProjections, pd.Series)
        super().__init__()

        for k, v in kwargs.items():
            if isinstance(v, dict):
                setattr(
                    self,
                    k,
                    DF_ICompanyEIProjections(
                        ei_metric=EI_Metric(v["ei_metric"]),
                        projections=v["projections"],
                    ),
                )
            elif isinstance(v, ICompanyEIProjections):
                setattr(self, k, DF_ICompanyEIProjections(icompany_ei_projections=v))
            elif isinstance(v, pd.Series):
                ei_metric = EI_Metric(str(v.dtype))
                if ei_metric.startswith("pint["):
                    ei_metric = ei_metric[5:-1]
                setattr(
                    self,
                    k,
                    DF_ICompanyEIProjections(ei_metric=ei_metric, projections=v),
                )
            elif isinstance(v, DF_ICompanyEIProjections) or v is None:
                setattr(self, k, v)
            else:
                # breakpoint()
                assert False
            # We could do a post-hoc validation here...

    def __getitem__(self, item):
        return getattr(self, item)

    def __str__(self):
        dict_items = {
            scope: getattr(self, scope).projections
            for scope in ["S1", "S2", "S1S2", "S3", "S1S2S3"]
            if getattr(self, scope) is not None
        }
        return str(pd.DataFrame.from_dict(dict_items))


class IProductionRealization(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    year: int
    value: Optional[ProductionQuantity] = None


class IEmissionRealization(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    year: int
    value: Optional[EmissionsQuantity] = None

    def __getitem__(self, item):
        return getattr(self, item)

    def __eq__(self, o):
        assert self.year == o.year
        if ITR.isna(self.value.m) and ITR.isna(o.value.m):
            return True
        return self.value == o.value

    def add(self, o):
        assert self.year == o.year
        return IEmissionRealization(
            year=self.year,
            value=self.value if ITR.isna(o.value.m) else self.value + o.value.to(self.value.units),
        )


class IHistoricEmissionsScopes(BaseModel):
    S1: List[IEmissionRealization]
    S2: List[IEmissionRealization]
    S1S2: List[IEmissionRealization]
    S3: List[IEmissionRealization]
    S1S2S3: List[IEmissionRealization]

    def __getitem__(self, item):
        return getattr(self, item)

    def __str__(self):
        dict_items = {
            scope: (
                lambda z: (
                    idx := z[0],
                    values := z[1],
                    pd.Series(PA_(np.asarray(values), dtype="Mt CO2e"), index=idx),
                )[-1]
            )(list(zip(*[(x.year, round(x.value.to("Mt CO2e").m, 4)) for x in getattr(self, scope)])))
            for scope in ["S1", "S2", "S1S2", "S3", "S1S2S3"]
            if getattr(self, scope) is not None
        }
        return str(pd.DataFrame.from_dict(dict_items))


class IEIRealization(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    year: int
    value: Optional[EI_Quantity] = None

    def __getitem__(self, item):
        return getattr(self, item)

    def __eq__(self, o):
        assert self.year == o.year
        if ITR.isna(self.value.m) and ITR.isna(o.value.m):
            return True
        return self.value == o.value

    def add(self, o):
        assert self.year == o.year
        return IEIRealization(
            year=self.year,
            value=self.value if ITR.isna(o.value.m) else self.value + o.value.to(self.value.units),
        )


class IHistoricEIScopes(BaseModel):
    S1: List[IEIRealization]
    S2: List[IEIRealization]
    S1S2: List[IEIRealization]
    S3: List[IEIRealization]
    S1S2S3: List[IEIRealization]

    def __getitem__(self, item):
        return getattr(self, item)

    def __str__(self):
        dict_items = {
            scope: (
                lambda z: (
                    idx := z[0],
                    values := z[1],
                    pd.Series(PA_(np.asarray(values), dtype=ei_metric), index=idx),
                )[-1]
            )(list(zip(*[(x.year, round(x.value.to(ei_metric).m, 4)) for x in getattr(self, scope)])))
            for scope in ["S1", "S2", "S1S2", "S3", "S1S2S3"]
            # Work-around for https://github.com/hgrecco/pint/issues/1687
            for ei_metric in [str(ureg.parse_units(getattr(self, scope).ei_metric))]
            if getattr(self, scope) is not None
        }
        return str(pd.DataFrame.from_dict(dict_items))


class IHistoricData(BaseModel):
    productions: Optional[List[IProductionRealization]] = None
    emissions: Optional[IHistoricEmissionsScopes] = None
    emissions_intensities: Optional[IHistoricEIScopes] = None


class ITargetData(BaseModel):
    netzero_year: Optional[int]
    target_type: Union[
        Literal["intensity"],
        Literal["absolute"],
        Literal["Intensity"],
        Literal["Absolute"],
    ]
    target_scope: EScope
    target_start_year: Optional[int] = None
    target_base_year: int
    target_end_year: int

    target_base_year_qty: float
    target_base_year_err: Optional[float] = None
    target_base_year_unit: str
    target_reduction_pct: float  # This is actually a fraction, not a percentage.  1.0 = complete reduction to zero.

    @model_validator(mode="before")
    def start_end_base_order(cls, v):
        if v["target_start_year"] < v["target_base_year"]:
            raise ValueError(
                f"Scope {v['target_scope']}: Target start year ({v['target_start_year']}) must be equal or greater than base year {v['target_base_year']}"
            )
        if v["target_end_year"] <= v["target_base_year"]:
            raise ValueError(
                f"Scope {v['target_scope']}: Target end year ({v['target_end_year']}) must be greater than base year {v['target_base_year']}"
            )
        if v["target_end_year"] <= v["target_start_year"]:
            raise ValueError(
                f"Scope {v['target_scope']}: Target end year ({v['target_end_year']}) must be greater than start year {v['target_start_year']}"
            )
        return v


class ICompanyData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    company_name: str
    company_id: str

    sector: str  # TODO: make SortableEnums
    region: str  # TODO: make SortableEnums
    # TemperatureScoreConfig.CONTROLS_CONFIG.target_probability is not company-specific,
    # while target_probability in ICompanyData is company-specific
    target_probability: float = np.nan

    target_data: Optional[List[ITargetData]] = None
    # IHistoric data can contain None values; need to convert to Quantified NaNs
    historic_data: Optional[IHistoricData] = None

    country: Optional[str] = None

    # Emissions typically use t CO2 for MWh/GJ and Mt CO2 for TWh/PJ
    emissions_metric: Optional[EmissionsMetric] = None
    production_metric: Optional[ProductionMetric] = None

    # These three instance variables match against financial data below, but are incomplete as historic_data and target_data
    base_year_production: Optional[ProductionQuantity] = None
    ghg_s1s2: Optional[EmissionsQuantity] = None
    ghg_s3: Optional[EmissionsQuantity] = None

    industry_level_1: Optional[str] = None
    industry_level_2: Optional[str] = None
    industry_level_3: Optional[str] = None
    industry_level_4: Optional[str] = None

    company_revenue: Optional[MonetaryQuantity] = None
    company_market_cap: Optional[MonetaryQuantity] = None
    company_enterprise_value: Optional[MonetaryQuantity] = None
    company_ev_plus_cash: Optional[MonetaryQuantity] = None
    company_total_assets: Optional[MonetaryQuantity] = None
    company_cash_equivalents: Optional[MonetaryQuantity] = None

    # Initialized later when we have benchmark information.  It is OK to initialize as None and fix later.
    # They will show up as {'S1S2': { 'projections': [ ... ] }}
    projected_targets: Optional[ICompanyEIProjectionsScopes] = None
    projected_intensities: Optional[ICompanyEIProjectionsScopes] = None

    # TODO: Do we want to do some sector inferencing here?

    def _sector_to_production_units(self, sector, region="Global"):
        sector_unit_dict = {
            "Electricity Utilities": {"North America": "MWh", "Global": "GJ"},
            "Gas Utilities": {"Global": "PJ"},
            "Utilities": {"Global": "PJ"},
            "Steel": {"Global": "t Steel"},
            "Aluminum": {"Global": "t Aluminum"},
            "Energy": {"Global": "PJ"},
            "Coal": {"Global": "t Coal"},
            "Oil": {"Global": "bbl/d"},
            "Gas": {"Global": "bcm"},
            "Oil & Gas": {"Global": "PJ"},
            "Autos": {"Global": "pkm"},
            "Trucking": {"Global": "tkm"},
            "Cement": {"Global": "t Cement"},
            "Construction Buildings": {"Global": "billion USD"},
            "Residential Buildings": {"Global": "billion m**2"},  # Should it be 'built m**2' ?
            "Commercial Buildings": {"Global": "billion m**2"},  # Should it be 'built m**2' ?
            "Textiles": {"Global": "billion USD"},
            "Chemicals": {"Global": "billion USD"},
            "Chemicals": {"Global": "billion USD"},
            "Pharmaceuticals": {"Global": "billion USD"},
            "Ag Chem": {"Global": "billion USD"},
            "Consumer Products": {"Global": "billion USD"},
            "Fiber & Rubber": {"Global": "billion USD"},
            "Petrochem & Plastics": {"Global": "billion USD"},
        }
        units = None
        if sector_unit_dict.get(sector):
            region_unit_dict = sector_unit_dict[sector]
            if region_unit_dict.get(region):
                units = region_unit_dict[region]
            else:
                units = region_unit_dict["Global"]
        else:
            raise ValueError(f"No source of production metrics for {self.company_name}")
        return units

    def _get_base_realization_from_historic(self, realized_values: List[BaseModel], metric, base_year=None):
        valid_realizations = [rv for rv in realized_values if not ITR.isna(rv.value)]
        if not valid_realizations:
            retval = realized_values[0].model_copy()
            retval.year = None
            return retval
        valid_realizations.sort(key=lambda x: x.year, reverse=True)
        if base_year and valid_realizations[0].year != base_year:
            retval = realized_values[0].copy()
            retval.year = base_year
            # FIXME: Unless and until we accept uncertainties as input, rather than computed data, we don't need to make this a UFloat here
            retval.value = PintType(metric).na_value
            return retval
        return valid_realizations[0]

    def _normalize_historic_data(
        self,
        historic_data: IHistoricData,
        production_metric: ProductionMetric,
        emissions_metric: EmissionsMetric,
    ) -> IHistoricData:
        def _normalize(value, metric):
            if value is not None:
                if value.u == metric:
                    return value
                if ITR.isna(value):
                    return PintType(metric).na_value
                # We've pre-conditioned metric so don't need to work around https://github.com/hgrecco/pint/issues/1687
                return value.to(metric)
            return PintType(metric).na_value

        if historic_data is None:
            return None

        if historic_data.productions:
            # Work-around for https://github.com/hgrecco/pint/issues/1687
            production_metric = ureg.parse_units(production_metric)  # Catch things like '$'
            historic_data.productions = [
                IProductionRealization(year=p.year, value=_normalize(p.value, production_metric))
                for p in historic_data.productions
            ]
        # Work-around for https://github.com/hgrecco/pint/issues/1687
        # emissions_metric = ureg(emissions_metric).u
        ei_metric = ureg.parse_units(f"{emissions_metric} / ({production_metric})")
        for scope_name in EScope.get_scopes():
            if historic_data.emissions:
                setattr(
                    historic_data.emissions,
                    scope_name,
                    [
                        IEmissionRealization(year=p.year, value=_normalize(p.value, emissions_metric))
                        for p in historic_data.emissions[scope_name]
                    ],
                )
            if historic_data.emissions_intensities:
                setattr(
                    historic_data.emissions_intensities,
                    scope_name,
                    [
                        IEIRealization(year=p.year, value=_normalize(p.value, ei_metric))
                        for p in historic_data.emissions_intensities[scope_name]
                    ],
                )
        return historic_data

    def __init__(
        self,
        emissions_metric=None,
        production_metric=None,
        base_year_production=None,
        ghg_s1s2=None,
        ghg_s3=None,
        target_data=None,
        historic_data=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            emissions_metric=emissions_metric,
            production_metric=production_metric,
            base_year_production=base_year_production,
            ghg_s1s2=ghg_s1s2,
            ghg_s3=ghg_s3,
            target_data=target_data,
            historic_data=historic_data,
            *args,
            **kwargs,
        )
        # In-bound parameters are JSON (str, int, float, dict), which are converted to models by __super__ and stored as instance variables
        if production_metric is None:
            units = self._sector_to_production_units(self.sector, self.region)
            self.production_metric = ProductionMetric(units)
            if emissions_metric is None:
                self.emissions_metric = EmissionsMetric("t CO2")
        elif emissions_metric is None:
            if str(self.production_metric) in [
                "TWh",
                "PJ",
                "Mt Steel",
                "megaFe_ton",
                "mmboe",
            ]:
                self.emissions_metric = EmissionsMetric("Mt CO2")
            else:
                self.emissions_metric = EmissionsMetric("t CO2")
            # TODO: Should raise a warning here

        # This is only a partial initialization
        if self.historic_data is None:
            return
        self.historic_data = self._normalize_historic_data(
            self.historic_data, self.production_metric, self.emissions_metric
        )
        base_year = None
        if self.base_year_production:
            pass
        # Right now historic_data comes in via template.py ESG data
        elif self.historic_data.productions:
            # TODO: This is a hack to get things going.
            base_realization = self._get_base_realization_from_historic(
                self.historic_data.productions, self.production_metric, base_year
            )
            base_year = base_realization.year
            self.base_year_production = base_realization.value
        else:
            logger.warning(f"missing historic data for base_year_production for {self.company_name}")
            self.base_year_production = PintType(self.production_metric).na_value
        if self.ghg_s1s2 is None and self.historic_data.emissions:
            if self.historic_data.emissions.S1S2:
                base_realization = self._get_base_realization_from_historic(
                    self.historic_data.emissions.S1S2, self.emissions_metric, base_year
                )
                base_year = base_year or base_realization.year
                self.ghg_s1s2 = base_realization.value
            elif self.historic_data.emissions.S1 and self.historic_data.emissions.S2:
                base_realization_s1 = self._get_base_realization_from_historic(
                    self.historic_data.emissions.S1, self.emissions_metric, base_year
                )
                base_realization_s2 = self._get_base_realization_from_historic(
                    self.historic_data.emissions.S2, self.emissions_metric, base_year
                )
                base_year = base_year or base_realization_s1.year
                if base_realization_s1.value is not None and base_realization_s2.value is not None:
                    self.ghg_s1s2 = base_realization_s1.value + base_realization_s2.value
        if self.ghg_s1s2 is None and self.historic_data.emissions_intensities:
            intensity_metric = ureg.parse_units(f"({self.emissions_metric}) / ({self.production_metric})")
            if self.historic_data.emissions_intensities.S1S2:
                base_realization = self._get_base_realization_from_historic(
                    self.historic_data.emissions_intensities.S1S2,
                    intensity_metric,
                    base_year,
                )
                base_year = base_year or base_realization.year
                if base_realization.value is not None:
                    self.ghg_s1s2 = base_realization.value * self.base_year_production
            elif self.historic_data.emissions_intensities.S1 and self.historic_data.emissions_intensities.S2:
                base_realization_s1 = self._get_base_realization_from_historic(
                    self.historic_data.emissions_intensities.S1,
                    intensity_metric,
                    base_year,
                )
                base_realization_s2 = self._get_base_realization_from_historic(
                    self.historic_data.emissions_intensities.S2,
                    intensity_metric,
                    base_year,
                )
                base_year = base_year or base_realization_s1.year
                if base_realization_s1.value is not None and base_realization_s2.value is not None:
                    self.ghg_s1s2 = (base_realization_s1.value + base_realization_s2.value) * self.base_year_production
            else:
                raise ValueError(f"missing S1S2 historic intensity data for {self.company_name}")
        if self.ghg_s1s2 is None:
            raise ValueError(
                f"missing historic emissions or intensity data to calculate ghg_s1s2 for {self.company_name}"
            )
        if self.ghg_s3 is None and self.historic_data.emissions and self.historic_data.emissions.S3:
            base_realization_s3 = self._get_base_realization_from_historic(
                self.historic_data.emissions.S3, self.emissions_metric, base_year
            )
            self.ghg_s3 = base_realization_s3.value
        if self.ghg_s3 is None and self.historic_data.emissions_intensities:
            if self.historic_data.emissions_intensities.S3:
                intensity_metric = ureg.parse_units(f"({self.emissions_metric}) / ({self.production_metric})")
                base_realization_s3 = self._get_base_realization_from_historic(
                    self.historic_data.emissions_intensities.S3,
                    intensity_metric,
                    base_year,
                )
                if base_realization_s3.value is not None:
                    self.ghg_s3 = base_realization_s3.value * self.base_year_production


# These aggregate terms are all derived from the benchmark being used
class ICompanyAggregates(ICompanyData):
    cumulative_budget: Optional[EmissionsQuantity] = None
    cumulative_scaled_budget: Optional[EmissionsQuantity] = None
    cumulative_trajectory: Optional[EmissionsQuantity] = None
    cumulative_target: Optional[EmissionsQuantity] = None
    benchmark_temperature: Optional[Quantity_type("delta_degC")] = None
    benchmark_global_budget: Optional[EmissionsQuantity] = None
    scope: Optional[EScope] = None

    # The first year that cumulative_projections exceeds the 2050 cumulative_budget
    trajectory_exceedance_year: Optional[int] = None
    target_exceedance_year: Optional[int] = None

    # projected_production is computed but never saved, so computed at least 2x: initialiation/projection and cumulative budget
    # projected_targets: Optional[ICompanyEIProjectionsScopes] = None
    # projected_intensities: Optional[ICompanyEIProjectionsScopes] = None

    # Custom validator here
    @field_validator("trajectory_exceedance_year", "target_exceedance_year")
    def val_exceedance_year(cls, v):
        if isinstance(v, int):
            return v
        if pd.isna(v):
            return None
        raise ValueError(f"{v} is not compatible with Int64 dtype")

    @classmethod
    def from_ICompanyData(cls, super_instance, scope_company_data):
        """
        Fast way to add instance variables to a pre-validated SUPER_INSTANCE
        SCOPE_COMPANY_DATA is the dictionary of the new values we want to add...for this one company
        """
        # FIXME: Would love to know how to run these automatically...
        EmissionsQuantity(scope_company_data["cumulative_budget"])
        EmissionsQuantity(scope_company_data["cumulative_scaled_budget"])
        if not ITR.isna(scope_company_data["cumulative_trajectory"]):
            EmissionsQuantity(scope_company_data["cumulative_trajectory"])
        if not ITR.isna(scope_company_data["cumulative_target"]):
            EmissionsQuantity(scope_company_data["cumulative_target"])
        if not Q_(scope_company_data["benchmark_temperature"]).is_compatible_with(ureg("delta_degC")):
            raise ValueError(
                f"benchmark temperature {scope_company_data['benchmark_temperature']} is not compatible with delta_degC"
            )
        else:
            scope_company_data["benchmark_temperature"] = Q_(scope_company_data["benchmark_temperature"])
        EmissionsQuantity(scope_company_data["benchmark_global_budget"])
        if not isinstance(scope_company_data["scope"], EScope):
            raise ValueError(f"scope {scope_company_data['scope']} is not a valid scope")
        if not ITR.isna(scope_company_data["trajectory_exceedance_year"]) and not isinstance(
            scope_company_data["trajectory_exceedance_year"], int
        ):
            raise ValueError(
                f"scope {scope_company_data['trajectory_exceedance_year']} is not a valid trajectory exceedance year value"
            )
        if not ITR.isna(scope_company_data["target_exceedance_year"]) and not isinstance(
            scope_company_data["target_exceedance_year"], int
        ):
            raise ValueError(
                f"scope {scope_company_data['target_exceedance_year']} is not a valid target exceedance year value"
            )
        # ...while not re-running any validation on super_instnace
        return cls.model_construct(**scope_company_data, **super_instance.__dict__)



from abc import ABC
from enum import Enum
import warnings  # needed until apply behaves better with Pint quantities in arrays

from typing import Type

from pint import Quantity
from pint_pandas import PintType

import numpy as np
import pandas as pd
import pint
import pint_pandas

import ITR
from .data.osc_units import asPintSeries, Q_, PA_
from .configs import PortfolioAggregationConfig, ColumnsConfig, LoggingConfig

import logging

logger = logging.getLogger(__name__)
LoggingConfig.add_config_to_logger(logger)

from .interfaces import EScope


class PortfolioAggregationMethod(Enum):
    """
    The portfolio aggregation method determines how the temperature scores for the individual companies are aggregated
    into a single portfolio score.
    """

    WATS = "WATS"
    TETS = "TETS"
    MOTS = "MOTS"
    EOTS = "EOTS"
    ECOTS = "ECOTS"
    AOTS = "AOTS"
    ROTS = "ROTS"

    @staticmethod
    def is_emissions_based(method: "PortfolioAggregationMethod") -> bool:
        """
        Check whether a given method is emissions-based (i.e. it uses the emissions to calculate the aggregation).

        :param method: The method to check
        :return:
        """
        return method in [
            PortfolioAggregationMethod.MOTS,
            PortfolioAggregationMethod.EOTS,
            PortfolioAggregationMethod.ECOTS,
            PortfolioAggregationMethod.AOTS,
            PortfolioAggregationMethod.ROTS,
        ]

    @staticmethod
    def get_value_column(method: "PortfolioAggregationMethod", column_config: Type[ColumnsConfig]) -> str:
        map_value_column = {
            PortfolioAggregationMethod.MOTS: column_config.COMPANY_MARKET_CAP,
            PortfolioAggregationMethod.EOTS: column_config.COMPANY_ENTERPRISE_VALUE,
            PortfolioAggregationMethod.ECOTS: column_config.COMPANY_EV_PLUS_CASH,
            PortfolioAggregationMethod.AOTS: column_config.COMPANY_TOTAL_ASSETS,
            PortfolioAggregationMethod.ROTS: column_config.COMPANY_REVENUE,
            # The test case tells us these should be correct
            PortfolioAggregationMethod.WATS: column_config.COMPANY_MARKET_CAP,
            PortfolioAggregationMethod.TETS: column_config.COMPANY_MARKET_CAP,
        }

        try:
            return map_value_column[method]
        except KeyError:
            logger.warning(
                f"method '{method}' not found (type({method}) = {type(method)}; defaulting to COMPANY_MARKET_CAP"
            )
        return column_config.COMPANY_MARKET_CAP


class PortfolioAggregation(ABC):
    """
    This class is a base class that provides portfolio aggregation calculation.

    :param config: A class defining the constants that are used throughout this class. This parameter is only required
                    if you'd like to overwrite a constant. This can be done by extending the PortfolioAggregationConfig
                    class and overwriting one of the parameters.
    """

    def __init__(self, config: Type[PortfolioAggregationConfig] = PortfolioAggregationConfig):
        self.c = config

    def _check_column(self, data: pd.DataFrame, column: str):
        """
        Check if a certain column is filled for all companies. If not log an error.
        The aggregation treats missing values as zeroes.

        :param data: The data to check
        :param column: The column to check
        :return:
        """
        missing_data = data[ITR.isna(data[column])][self.c.COLS.COMPANY_NAME].unique()
        if len(missing_data):
            logger.error(f"The value for {column} is missing for the following companies: {', '.join(missing_data)}")

    def _calculate_aggregate_score(
        self,
        data: pd.DataFrame,
        input_column: str,
        portfolio_aggregation_method: PortfolioAggregationMethod,
    ) -> pd.Series:
        """
        Aggregate the scores in a given column based on a certain portfolio aggregation method.

        :param data: The data to run the calculations on
        :param input_column: The input column (containing the scores)
        :param portfolio_aggregation_method: The method to use
        :return: The aggregates score as a pd.Series
        """
        # Used to test against data[input_column].dtype.kind in ['f', 'i']
        assert isinstance(data[input_column].dtype, PintType)
        if portfolio_aggregation_method == PortfolioAggregationMethod.WATS:
            # Used to test against data[self.c.COLS.INVESTMENT_VALUE].dtype.kind in ['f', 'i']
            assert isinstance(data[self.c.COLS.INVESTMENT_VALUE].dtype, PintType)
            total_investment_weight = data[self.c.COLS.INVESTMENT_VALUE].sum()
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    # See https://github.com/hgrecco/pint-pandas/issues/114
                    weights_series = data[input_column] * data[self.c.COLS.INVESTMENT_VALUE] / total_investment_weight
                    return weights_series
            except ZeroDivisionError:
                raise ValueError("The portfolio weight is not allowed to be zero")

        # Total emissions weighted temperature score (TETS)
        elif portfolio_aggregation_method == PortfolioAggregationMethod.TETS:
            use_S1S2 = data[self.c.COLS.SCOPE].isin([EScope.S1, EScope.S2, EScope.S1S2, EScope.S1S2S3])
            use_S3 = data[self.c.COLS.SCOPE].isin([EScope.S3, EScope.S1S2S3])
            assert isinstance(data[self.c.COLS.GHG_SCOPE12].dtype, PintType)
            assert isinstance(data[self.c.COLS.GHG_SCOPE3].dtype, PintType)
            if use_S3.any():
                self._check_column(data, self.c.COLS.GHG_SCOPE3)
                use_S3 = use_S3 & ~ITR.isna(data[self.c.COLS.GHG_SCOPE3])
            if use_S1S2.any():
                self._check_column(data, self.c.COLS.GHG_SCOPE12)
                use_S1S2 = use_S1S2 & ~ITR.isna(data[self.c.COLS.GHG_SCOPE12])
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    # Calculate the total emissions of all companies.
                    # https://github.com/pandas-dev/pandas/issues/50564 explains why we need fillna(0) to make sum work
                    emissions = (
                        asPintSeries(data.loc[use_S1S2, self.c.COLS.GHG_SCOPE12]).fillna(0).sum()
                        + asPintSeries(data.loc[use_S3, self.c.COLS.GHG_SCOPE3]).fillna(0).sum()
                    )
                    # See https://github.com/hgrecco/pint-pandas/issues/130
                    weights_dtype = f"pint[{emissions.u}]"
                    # df_z works around fact that we cannot just insert quantities willy-nilly using Pandas where function
                    df_z = pd.Series(
                        PA_(np.zeros(len(data.index)), dtype="Mt CO2e"),
                        index=data.index,
                    )
                    weights_series = (
                        (
                            data[self.c.COLS.GHG_SCOPE12].where(use_S1S2, df_z)
                            + data[self.c.COLS.GHG_SCOPE3].where(use_S3, df_z)
                        ).astype(weights_dtype)
                        / emissions
                        * data[input_column]
                    )
                    return weights_series

            except ZeroDivisionError:
                raise ValueError("The total emissions should be higher than zero")

        elif PortfolioAggregationMethod.is_emissions_based(portfolio_aggregation_method):
            # These four methods only differ in the way the company is valued.
            value_column = PortfolioAggregationMethod.get_value_column(portfolio_aggregation_method, self.c.COLS)
            # Used to check data[value_column].dtype.kind in ['f', 'i']
            assert isinstance(data[value_column].dtype, PintType)

            # Calculate the total owned emissions of all companies
            try:
                self._check_column(data, self.c.COLS.INVESTMENT_VALUE)
                self._check_column(data, value_column)
                use_S1S2 = data[self.c.COLS.SCOPE].isin([EScope.S1, EScope.S2, EScope.S1S2, EScope.S1S2S3])
                use_S3 = data[self.c.COLS.SCOPE].isin([EScope.S3, EScope.S1S2S3])
                if use_S1S2.any():
                    self._check_column(data, self.c.COLS.GHG_SCOPE12)
                    use_S1S2 = use_S1S2 & ~ITR.isna(data[self.c.COLS.GHG_SCOPE12])
                if use_S3.any():
                    self._check_column(data, self.c.COLS.GHG_SCOPE3)
                    use_S3 = use_S3 & ~ITR.isna(data[self.c.COLS.GHG_SCOPE3])
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    df_z = pd.Series(
                        PA_(np.zeros(len(data.index)), dtype="Mt CO2e"),
                        index=data.index,
                    )
                    data[self.c.COLS.OWNED_EMISSIONS] = (data[self.c.COLS.INVESTMENT_VALUE] / data[value_column]) * (
                        data[self.c.COLS.GHG_SCOPE12].where(use_S1S2, df_z)
                        + data[self.c.COLS.GHG_SCOPE3].where(use_S3, df_z)
                    ).astype("pint[Mt CO2e]")
            except ZeroDivisionError:
                raise ValueError("To calculate the aggregation, the {} column may not be zero".format(value_column))

            try:
                # Calculate the MOTS value per company
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    assert isinstance(data[self.c.COLS.OWNED_EMISSIONS].dtype, PintType)
                    owned_emissions = asPintSeries(data[self.c.COLS.OWNED_EMISSIONS])
                    # https://github.com/pandas-dev/pandas/issues/50564 explains why we need fillna(0) to make sum work
                    total_emissions = owned_emissions.fillna(0).sum()
                    result = data[input_column] * owned_emissions / total_emissions
                return result
            except ZeroDivisionError:
                raise ValueError("The total owned emissions can not be zero")
        else:
            raise ValueError("The specified portfolio aggregation method is invalid")



import warnings  # needed until apply behaves better with Pint quantities in arrays
from typing import Type, List, Optional
import pandas as pd
import numpy as np

from pydantic import BaseModel, ValidationError

from . import ureg, Q_
from ITR.data.osc_units import (
    BenchmarkMetric,
    EmissionsQuantity,
    EI_Quantity,
    Quantity_type,
)

from ITR.configs import (
    ColumnsConfig,
    TemperatureScoreConfig,
    VariablesConfig,
    TabsConfig,
    ProjectionControls,
    LoggingConfig,
)

import logging

logger = logging.getLogger(__name__)
LoggingConfig.add_config_to_logger(logger)

from ITR.data.base_providers import (
    BaseCompanyDataProvider,
    BaseProviderProductionBenchmark,
    BaseProviderIntensityBenchmark,
)
from ITR.interfaces import (
    ICompanyData,
    ICompanyEIProjection,
    EScope,
    IEIBenchmarkScopes,
    IProductionBenchmarkScopes,
    IBenchmark,
    IBenchmarks,
    IHistoricEmissionsScopes,
    IProductionRealization,
    ProductionQuantity,
    IHistoricEIScopes,
    IHistoricData,
    ITargetData,
    IEmissionRealization,
    IEIRealization,
    UProjection,
    IProjection,
)


# Excel spreadsheets don't have units elaborated, so we translate sectors to units
# FIXME: this is now out of data with our much better JSON-based benchmark data
sector_to_production_metric = {
    "Electricity Utilities": "GJ",
    "Steel": "Fe_ton",
    "Oil & Gas": "boe",
    "Autos": "pkm",
}
sector_to_intensity_metric = {
    "Electricity Utilities": "t CO2/MWh",
    "Steel": "t CO2/Fe_ton",
    "Oil & Gas": "kg CO2/boe",
    "Autos": "g CO2/pkm",
}

# TODO: Force validation for excel benchmarks

# Utils functions:


def convert_dimensionless_benchmark_excel_to_model(
    df_excel: dict, sheetname: str, column_name_sector: str, column_name_region: str
) -> IBenchmarks:
    """
    Converts excel into IBenchmarks
    :param df_excel: dictionary with a pd.DataFrame for each key representing a sheet of an Excel file
    :param sheetname: name of Excel file sheet to convert
    :param column_name_sector: name of sector
    :param column_name_region: name of region
    :return: IBenchmarks instance (list of IBenchmark)
    """
    try:
        df_sheet = df_excel[sheetname]
    except KeyError:
        logger.error(f"Sheet {sheetname} not in benchmark Excel file.")
        raise

    df_production = df_sheet.reset_index(drop=True).set_index(
        [column_name_sector, column_name_region, "benchmark_metric", "scope"]
    )

    result = []
    # FIXME: More pythonic to convert DF to dict and convert dict to Model
    for index, row in df_production.iterrows():
        bm = IBenchmark(
            sector=index[0],
            region=index[1],
            benchmark_metric=BenchmarkMetric("dimensionless"),
            projections_nounits=[UProjection(year=int(k), value=float(v)) for k, v in row.items()],
        )
        result.append(bm)
    return IBenchmarks(benchmarks=result)


def convert_benchmarks_ei_excel_to_model(
    df_excel: pd.DataFrame,
    sheetname: str,
    column_name_sector: str,
    column_name_region: str,
    benchmark_temperature,
    benchmark_global_budget,
    is_AFOLU_included,
) -> IEIBenchmarkScopes:
    """
    Converts excel into IBenchmarks
    :param excal_path: file path to excel
    :return: IEIBenchmarkScopes instance
    """
    df_ei_bms = (
        df_excel[sheetname]
        .reset_index(drop=True)
        .set_index([column_name_sector, column_name_region, "benchmark_metric", "scope"])
    )
    bm_dict = {scope_name: [] for scope_name in EScope.get_scopes()}
    bm_dict["benchmark_temperature"] = benchmark_temperature
    bm_dict["benchmark_global_budget"] = benchmark_global_budget
    bm_dict["is_AFOLU_included"] = is_AFOLU_included
    # FIXME: More pythonic to convert DF to dict and convert dict to Model
    for index, row in df_ei_bms.iterrows():
        bm = IBenchmark(
            sector=index[0],
            region=index[1],
            benchmark_metric=index[2],
            projections=[IProjection(year=int(k), value=Q_(float(v), index[2])) for k, v in row.items()],
        )
        bm_dict[index[3]].append(bm)
    for scope_name in EScope.get_scopes():
        bm_dict[scope_name] = IBenchmarks(benchmarks=bm_dict[scope_name])
    return IEIBenchmarkScopes(**bm_dict)


class ExcelProviderProductionBenchmark(BaseProviderProductionBenchmark):
    def __init__(self, excel_path: str, column_config: Type[ColumnsConfig] = ColumnsConfig):
        """
        Overrices BaseProvider and provides an interfaces for excel the excel template
        :param excel_path: file path to excel
        :param column_config: An optional ColumnsConfig object containing relevant variable names
        """
        self.benchmark_excel = pd.read_excel(excel_path, sheet_name=None, skiprows=0)
        for sheetname, df in self.benchmark_excel.items():
            # This fills down the sector information for different regions
            self.benchmark_excel[sheetname] = df.ffill()
        self._convert_excel_to_model = convert_dimensionless_benchmark_excel_to_model
        production_bms = self._convert_excel_to_model(
            self.benchmark_excel,
            TabsConfig.PROJECTED_PRODUCTION,
            column_config.SECTOR,
            column_config.REGION,
        )
        super().__init__(IProductionBenchmarkScopes(AnyScope=production_bms), column_config)

    def _get_projected_production(self, scope: EScope = EScope.S1S2) -> pd.DataFrame:
        """
        interface from excel file and internally used DataFrame
        :param scope:
        :return:
        """
        # df = self.benchmark_excel[TabsConfig.PROJECTED_PRODUCTION].drop(columns='benchmark_metric')
        # df.loc[:, 'scope'] = df.scope.map(lambda x: EScope[x])
        # df.set_index([self.column_config.SECTOR, self.column_config.REGION, self.column_config.SCOPE], inplace=True)
        # df_partial_pp = df.add(1).cumprod(axis=1).astype('pint[dimensionless]')
        # return df_partial_pp
        return self._prod_df


class ExcelProviderIntensityBenchmark(BaseProviderIntensityBenchmark):
    def __init__(
        self,
        excel_path: str,
        benchmark_temperature: Quantity_type("delta_degC"),
        benchmark_global_budget: EmissionsQuantity,
        is_AFOLU_included: bool,
        column_config: Type[ColumnsConfig] = ColumnsConfig,
    ):
        self.benchmark_excel = pd.read_excel(excel_path, sheet_name=None, skiprows=0)
        for sheetname, df in self.benchmark_excel.items():
            # This fills down the sector information for different regions
            self.benchmark_excel[sheetname] = df.ffill()
        self._convert_excel_to_model = convert_benchmarks_ei_excel_to_model
        ei_bm_scopes = self._convert_excel_to_model(
            self.benchmark_excel,
            TabsConfig.PROJECTED_EI,
            column_config.SECTOR,
            column_config.REGION,
            benchmark_temperature,
            benchmark_global_budget,
            is_AFOLU_included,
        )
        super().__init__(ei_bm_scopes, column_config)


# FIXME: Should we merge with TemplateProviderCompany and just use a different excel input method
# for this "simple" case?
class ExcelProviderCompany(BaseCompanyDataProvider):
    """
    Data provider skeleton for CSV files. This class serves primarily for testing purposes only!

    :param excel_path: A path to the Excel file with the company data
    :param column_config: An optional ColumnsConfig object containing relevant variable names
    """

    def __init__(
        self,
        excel_path: str,
        column_config: Type[ColumnsConfig] = ColumnsConfig,
        projection_controls: Type[ProjectionControls] = ProjectionControls,
    ):
        self.projection_controls = projection_controls
        self._companies = self._convert_from_excel_data(excel_path)
        self.historic_years = None
        super().__init__(self._companies, column_config, projection_controls)

    def _check_company_data(self, company_tabs: dict) -> None:
        """
        Checks if the company data excel contains the data in the right format

        :return: None
        """
        required_tabs = {TabsConfig.FUNDAMENTAL, TabsConfig.PROJECTED_TARGET}
        optional_tabs = {TabsConfig.PROJECTED_EI, TabsConfig.HISTORIC_DATA}
        missing_tabs = (required_tabs | optional_tabs).difference(set(company_tabs))
        if missing_tabs.intersection(required_tabs):
            logger.error(f"Tabs {required_tabs} are required.")
            raise ValueError(f"Tabs {required_tabs} are required.")
        if optional_tabs.issubset(missing_tabs):
            logger.error(f"Either of the tabs {optional_tabs} is required.")
            raise ValueError(f"Either of the tabs {optional_tabs} is required.")

    def _convert_from_excel_data(self, excel_path: str) -> List[ICompanyData]:
        """
        Converts the Excel template to list of ICompanyDta objects. All dataprovider features will be inhereted from
        Base
        :param excel_path: file path to excel file
        :return: List of ICompanyData objects
        """
        company_data = pd.read_excel(excel_path, sheet_name=None, skiprows=0)
        self._check_company_data(company_data)

        df_fundamentals = company_data[TabsConfig.FUNDAMENTAL].set_index(ColumnsConfig.COMPANY_ID)
        df_fundamentals[ColumnsConfig.PRODUCTION_METRIC] = df_fundamentals[ColumnsConfig.SECTOR].map(
            sector_to_production_metric
        )
        company_ids = df_fundamentals.index.unique().get_level_values(level=0).tolist()
        # _get_projection creates S1S2 data from S1+S2.  _get_historic_data must do the same to keep up.
        df_targets = self._get_projection(
            company_ids,
            company_data[TabsConfig.PROJECTED_TARGET],
            df_fundamentals[ColumnsConfig.PRODUCTION_METRIC],
        )
        if TabsConfig.PROJECTED_EI in company_data:
            df_ei = self._get_projection(
                company_ids,
                company_data[TabsConfig.PROJECTED_EI],
                df_fundamentals[ColumnsConfig.PRODUCTION_METRIC],
            )
        else:
            df_ei = None
        if TabsConfig.HISTORIC_DATA in company_data:
            df_historic = company_data[TabsConfig.HISTORIC_DATA].set_index(ColumnsConfig.COMPANY_ID)
            # DON'T update historic data
            if False:
                df_prods = df_historic[df_historic.variable == "Productions"].drop(columns=["variable", "scope"])
                df_s1s2 = (
                    df_historic[df_historic.variable == "Emissions"]
                    .groupby(by=["company_id", "variable"])
                    .sum()
                    .droplevel("variable")
                )
                df_s1s2_ei = (
                    df_historic[df_historic.variable == "Emissions Intensities"]
                    .groupby(by=["company_id", "variable"])
                    .sum()
                    .droplevel("variable")
                )
                em_not_ei = df_s1s2.index.difference(df_s1s2_ei.index).intersection(df_prods.index)
                ei_not_em = df_s1s2_ei.index.difference(df_s1s2.index).intersection(df_prods.index)
                df_ei_from_em = df_s1s2.loc[em_not_ei].div(df_prods.loc[em_not_ei])
                df_em_from_ei = df_s1s2_ei.loc[ei_not_em].mul(df_prods.loc[ei_not_em])
                df_s1s2["scope"] = df_s1s2_ei["scope"] = df_ei_from_em["scope"] = df_em_from_ei["scope"] = "S1S2"
                df_s1s2["variable"] = df_em_from_ei["variable"] = "Emissions"
                df_s1s2_ei["variable"] = df_ei_from_em["variable"] = "Emissions Intensities"
                df_historic = pd.concat([df_historic, df_s1s2, df_s1s2_ei, df_ei_from_em, df_em_from_ei])
            df_historic = df_historic.merge(
                df_fundamentals[ColumnsConfig.PRODUCTION_METRIC].rename("units"),
                left_index=True,
                right_index=True,
            )
            df_historic.loc[df_historic.variable == "Emissions", "units"] = "t CO2"
            # If you think the following line of code is ugly, please answer https://stackoverflow.com/q/74555323/1291237
            df_historic.loc[df_historic.variable == "Emissions Intensities", "units"] = df_historic.loc[
                df_historic.variable == "Emissions Intensities"
            ].units.map(lambda x: f"t CO2/({x})")
            df_historic = self._get_historic_data(company_ids, df_historic)
            company_data[TabsConfig.HISTORIC_DATA] = df_historic
        else:
            df_historic = None

        return self._company_df_to_model(df_fundamentals, df_targets, df_ei, df_historic)

    def _convert_series_to_projections(self, projections: pd.Series, ProjectionType: BaseModel) -> [IProjection]:
        """
        Converts a Pandas Series to a list of IProjection
        :param projections: Pandas Series with years as indices
        :return: List of IProjection objects
        """
        return [ProjectionType(year=y, value=v) for y, v in projections.items()]

    def _company_df_to_model(
        self,
        df_fundamentals: pd.DataFrame,
        df_targets: pd.DataFrame,
        df_ei: pd.DataFrame,
        df_historic: pd.DataFrame,
    ) -> List[ICompanyData]:
        """
        transforms target Dataframe into list of IDataProviderTarget instances
        :param df_fundamentals: pandas Dataframe with fundamental data
        :param df_targets: pandas Dataframe with targets
        :param df_ei: pandas Dataframe with emission intensities
        :return: A list containing the ICompanyData objects
        """
        # set NaN to None since NaN is float instance
        df_fundamentals = df_fundamentals.where(pd.notnull(df_fundamentals), None).replace({np.nan: None})

        companies_data_dict = df_fundamentals.to_dict(orient="index")
        model_companies: List[ICompanyData] = []
        for company_id, company_data in companies_data_dict.items():
            try:
                production_metric = sector_to_production_metric[company_data[ColumnsConfig.SECTOR]]
                intensity_metric = sector_to_intensity_metric[company_data[ColumnsConfig.SECTOR]]
                company_data[ColumnsConfig.PRODUCTION_METRIC] = production_metric
                company_data[ColumnsConfig.EMISSIONS_METRIC] = "t CO2"
                # pint automatically handles any unit conversions required

                v = df_fundamentals.loc[company_id][ColumnsConfig.GHG_SCOPE12]
                company_data[ColumnsConfig.GHG_SCOPE12] = Q_(np.nan if v is None else v, "t CO2")
                company_data[ColumnsConfig.BASE_YEAR_PRODUCTION] = (
                    company_data[ColumnsConfig.GHG_SCOPE12]
                    / df_ei.loc[company_id, :][self.projection_controls.BASE_YEAR]
                )
                v = df_fundamentals.loc[company_id][ColumnsConfig.GHG_SCOPE3]
                company_data[ColumnsConfig.GHG_SCOPE3] = Q_(np.nan if v is None else v, "t CO2")
                company_data[ColumnsConfig.PROJECTED_TARGETS] = {
                    "S1S2": {
                        "projections": self._convert_series_to_projections(
                            df_targets.loc[company_id, :], ICompanyEIProjection
                        ),
                        "ei_metric": intensity_metric,
                    }
                }
                company_data[ColumnsConfig.PROJECTED_EI] = {
                    "S1S2": {
                        "projections": self._convert_series_to_projections(
                            df_ei.loc[company_id, :], ICompanyEIProjection
                        ),
                        "ei_metric": intensity_metric,
                    }
                }

                fundamental_metrics = [
                    ColumnsConfig.COMPANY_MARKET_CAP,
                    ColumnsConfig.COMPANY_REVENUE,
                    ColumnsConfig.COMPANY_ENTERPRISE_VALUE,
                    ColumnsConfig.COMPANY_TOTAL_ASSETS,
                    ColumnsConfig.COMPANY_CASH_EQUIVALENTS,
                ]
                for col in fundamental_metrics:
                    company_data[col] = Q_(company_data[col], company_data[ColumnsConfig.COMPANY_CURRENCY])
                company_data[ColumnsConfig.COMPANY_EV_PLUS_CASH] = (
                    company_data[ColumnsConfig.COMPANY_ENTERPRISE_VALUE]
                    + company_data[ColumnsConfig.COMPANY_CASH_EQUIVALENTS]
                )
                if df_historic is not None:
                    company_data[TabsConfig.HISTORIC_DATA] = dict(
                        self._convert_historic_data(df_historic.loc[company_id, :])
                    )
                else:
                    company_data[TabsConfig.HISTORIC_DATA] = None

                # Put the index back into the dictionary so model builds correctly
                company_data[ColumnsConfig.COMPANY_ID] = company_id
                model_companies.append(ICompanyData.model_validate(company_data))
            except ValidationError as e:
                logger.warning(
                    f"EX {e}: (one of) the input(s) of company %s is invalid and will be skipped"
                    % company_data[ColumnsConfig.COMPANY_NAME]
                )
        return model_companies

    # Workaround for bug (https://github.com/pandas-dev/pandas/issues/20824) in Pandas where NaN are treated as zero
    def _np_sum(g):
        return np.sum(g.values)

    # FIXME: Need to deal with S3 emissions as well
    def _get_projection(
        self,
        company_ids: List[str],
        projections: pd.DataFrame,
        production_metric: pd.DataFrame,
    ) -> pd.DataFrame:
        """
        get the projected emission intensities for list of companies
        :param company_ids: list of company ids
        :param projections: Dataframe with listed projections per company
        :param production_metric: Dataframe with production_metric per company
        :return: series of projected emission intensities
        """
        projections = projections.reset_index().set_index(ColumnsConfig.COMPANY_ID)

        missing_companies = [company_id for company_id in company_ids if company_id not in projections.index]
        if missing_companies:
            error_message = f"Missing target or trajectory projections for companies with ID: {missing_companies}"
            logger.error(error_message)
            raise ValueError(error_message)

        projections = projections.loc[
            company_ids,
            range(
                self.projection_controls.BASE_YEAR,
                self.projection_controls.TARGET_YEAR + 1,
            ),
        ]
        # The following creates projections for S1+S2 from data giving only S1 and S2; Need to get historic_data to see that...
        # Due to bug (https://github.com/pandas-dev/pandas/issues/20824) in Pandas where NaN are treated as zero workaround below:
        projected_emissions_s1s2 = projections.groupby(level=0, sort=False).agg(
            ExcelProviderCompany._np_sum
        )  # add scope 1 and 2
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # See https://github.com/hgrecco/pint-pandas/issues/114
            projected_emissions_s1s2 = projected_emissions_s1s2.apply(
                lambda x: x.astype(f"pint[t CO2/({production_metric[x.name]})]"), axis=1
            )

        return projected_emissions_s1s2

    def _convert_target_data(self, target_data: pd.DataFrame) -> List[ITargetData]:
        """
        :param historic: historic production, emission and emission intensity data for a company
        :return: IHistoricData Pydantic object
        """
        target_data = target_data.rename(
            columns={
                "target_year": "target_end_year",
                "target_reduction_ambition": "target_reduction_pct",
            }
        )
        return [ITargetData(**td) for td in target_data.to_dict("records")]

    def _get_historic_data(self, company_ids: List[str], historic_data: pd.DataFrame) -> pd.DataFrame:
        """
        get the historic data for list of companies
        :param company_ids: list of company ids
        :param historic_data: Dataframe Productions, Emissions, and Emissions Intensities mixed together
        :return: historic data with unit attributes added on a per-element basis
        """
        self.historic_years = [column for column in historic_data.columns if type(column) == int]

        missing_ids = [company_id for company_id in company_ids if company_id not in historic_data.index]
        if missing_ids:
            error_message = f"Company ids missing in provided historic data: {missing_ids}"
            logger.error(error_message)
            raise ValueError(error_message)

        for year in self.historic_years:
            historic_data[year] = historic_data.apply(lambda x: f"{x[year]} {x.units}", axis=1)
        return historic_data.loc[company_ids]

    # In the following several methods, we implement SCOPE as STRING (used by Excel handlers)
    # so that the resulting scope dictionary can be used to pass values to named arguments
    def _convert_historic_data(self, historic: pd.DataFrame) -> IHistoricData:
        """
        :param historic: historic production, emission and emission intensity data for a company (already unitized)
        :return: IHistoricData Pydantic object
        """
        productions = historic.loc[historic[ColumnsConfig.VARIABLE] == VariablesConfig.PRODUCTIONS]
        emissions = historic.loc[historic[ColumnsConfig.VARIABLE] == VariablesConfig.EMISSIONS]
        emissions_intensities = historic.loc[historic[ColumnsConfig.VARIABLE] == VariablesConfig.EMISSIONS_INTENSITIES]
        hd = IHistoricData(
            productions=self._convert_to_historic_productions(productions),
            emissions=self._convert_to_historic_emissions(emissions),
            emissions_intensities=self._convert_to_historic_ei(emissions_intensities),
        )
        return hd

    # Note that for the three following functions, we pd.Series.squeeze() the results because it's just one year / one company
    def _convert_to_historic_emissions(self, emissions: pd.DataFrame) -> Optional[IHistoricEmissionsScopes]:
        """
        :param emissions: historic emissions data for a company
        :return: List of historic emissions per scope, or None if no data are provided
        """
        if emissions.empty:
            return None

        emissions_scopes = {}
        for scope_name in EScope.get_scopes():
            results = emissions.loc[emissions[ColumnsConfig.SCOPE] == scope_name]
            emissions_scopes[scope_name] = (
                []
                if results.empty
                else [
                    IEmissionRealization(year=year, value=EmissionsQuantity(results[year].squeeze()))
                    for year in self.historic_years
                ]
            )
        return IHistoricEmissionsScopes(**emissions_scopes)

    def _convert_to_historic_productions(self, productions: pd.DataFrame) -> Optional[List[IProductionRealization]]:
        """
        :param productions: historic production data for a company
        :return: A list containing historic productions, or None if no data are provided
        """
        if productions.empty:
            return None
        return [
            IProductionRealization(year=year, value=ProductionQuantity(productions[year].squeeze()))
            for year in self.historic_years
        ]

    def _convert_to_historic_ei(self, intensities: pd.DataFrame) -> Optional[IHistoricEIScopes]:
        """
        :param intensities: historic emission intensity data for a company
        :return: A list of historic emission intensities per scope, or None if no data are provided
        """
        if intensities.empty:
            return None

        intensities = intensities.copy()
        intensity_scopes = {}

        for scope_name in EScope.get_scopes():
            results = intensities.loc[intensities[ColumnsConfig.SCOPE] == scope_name]
            intensity_scopes[scope_name] = (
                []
                if results.empty
                else [
                    IEIRealization(year=year, value=EI_Quantity(results[year].squeeze()))
                    for year in self.historic_years
                ]
            )
        return IHistoricEIScopes(**intensity_scopes)
